//---------------------------------------------------------------------------
#include "Precompiled.h"
#pragma hdrstop

#include <sstream>
#include <iomanip>
#include <gimite/reg_manager.h>
#include <gimite/string_algo.h>
#include "Win32Common.h"
#include "VCLCommon.h"
#include "Maprof.h"
#include "Definitions.h"
#include "Backup.h"
using namespace std;
using namespace gimite;
using namespace boost::posix_time;
//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
TBackupUnit::TBackupUnit(TBackup* backup/*= NULL*/)
	: backup_(backup), time_(time_from_string("2000-01-01 00:00:00.000")){}
//---------------------------------------------------------------------------
void TBackupUnit::Backup(const TStringVector& paths)//obNAbv
{
	time_= second_clock::local_time();
	MakeName(paths);
	sizeFiles_= 0;
	string pathDestDir= CombineDirFile(backup_->BaseDirPath(), sName_);
	for (size_t i= 0; i<paths.size(); ++i){
		string bpath= BackupPath(paths[i]);
		bpathsFile_.push_back(bpath);
		string path= CombineDirFile(pathDestDir, bpath);
		CreateDir(ExtractFileDir(path.c_str()).c_str());
		CopyFile(paths[i].c_str(), path.c_str(), false);
		sizeFiles_+= GetCompressedFileSize(path.c_str(), NULL);
			//4Gȏ̃t@Cɂ͖Ή
	}
}

//---------------------------------------------------------------------------
void TBackupUnit::Delete()//obNAbv폜
{
	string pathDir= CombineDirFile(backup_->BaseDirPath(), sName_);
	for (size_t i= 0; i<bpathsFile_.size(); ++i)
		DeleteFile(CombineDirFile(pathDir, bpathsFile_[i]).c_str());
	RemoveDirAndSubDirs(pathDir);
}

//---------------------------------------------------------------------------
void TBackupUnit::ReadMembers(gimite::TGTSNode* node)
{
	node->Read("Time", time_);
	node->Read("Name", sName_);
	node->Read("Size", sizeFiles_);
	node->Read("BaseDir", pathBaseDir_);
	string s;
	bpathsFile_.clear();
	while (node->Read("File", s))
		bpathsFile_.push_back(s);
}

//---------------------------------------------------------------------------
void TBackupUnit::WriteMembers(gimite::TGTSNode* node)const
{
	node->Write("Time", time_);
	node->Write("Name", sName_);
	node->Write("Size", sizeFiles_);
	node->Write("BaseDir", pathBaseDir_);
	for (size_t i= 0; i<bpathsFile_.size(); ++i)
		node->Write("File", bpathsFile_[i]);
}

//---------------------------------------------------------------------------
void TBackupUnit::MakeName(const TStringVector& paths)//O
{
	reg_manager rm;
	pathBaseDir_= IncludingDir(paths);
	string s= ExtractFileName(pathBaseDir_.c_str()).c_str();
	if (s.empty()){
		s= pathBaseDir_;
		rm.replace(s, "[\\/;,:*?\"<>|]", "_", "g");
	}
	ostringstream ss;
	ss<<setfill('0')<<setw(4)<<time_.date().year()
		<<setw(2)<<time_.date().month()
		<<setw(2)<<time_.date().day()<<'_'<<s<<'_';
	string pathBase2= CombineDirFile(backup_->BaseDirPath(), "");
	int i= 1;
	while (GetPathKind(pathBase2+ss.str()+to_str(i))!=pkInvalid) ++i;
	sName_= ss.str()+to_str(i);
}

//---------------------------------------------------------------------------
std::string TBackupUnit::BackupPath(const std::string& path)
	//̃pXɑΉobNAbvpX
{
	reg_manager rm;
	string s= RelativePath(path, pathBaseDir_);
	rm.replace(s, "[/;,:*?\"<>|]", "_", "g");
	rm.replace(s, "\\\\\\\\", "", "g");
	return s;
}

//---------------------------------------------------------------------------
void TBackupUnit::CreateDir(const std::string& path)
	//tH_Ƃ̑SĂ̑ctH_
{
	if (path!=ExtractFileDrive(path.c_str()).c_str()
			&& path!=ExtractFileDir(path.c_str()).c_str()){
		CreateDir(ExtractFileDir(path.c_str()).c_str());
		::CreateDirectory(path.c_str(), NULL);
	}
}

//---------------------------------------------------------------------------
bool TBackupUnit::RemoveDirAndSubDirs(const std::string& path)
	//tH_Ƃ̑SẴTutH_폜
{
	ForEachSubDir(path, &TBackupUnit::RemoveDirAndSubDirs);
	RemoveDirectory(path.c_str());
	return true;
}

//---------------------------------------------------------------------------
TBackup::TBackup()
	: pathBaseDir_((ExtractFilePath(Application->ExeName)+"Backup").c_str()),
	sizeMax_(3*1024*1024), tmdDeletable_(hours(0))
{
}

//---------------------------------------------------------------------------
void TBackup::Backup(const TStringVector& paths)//obNAbv
{
	DWORD sizeUnit= FilesSize(paths);
	if (sizeUnit>=sizeMax_
			&& (onBackupExceedingUnit_.empty()
			|| !onBackupExceedingUnit_(sizeUnit)))
		return;//񂾂ŃTCY̏𒴂̂ŃobNAbvȂ

	units_.push_back(TBackupUnit(this));
	units_.back().Backup(paths);

	//TCYI[o̕폜
	DWORD size= TotalSize();
	ptime timeCur= second_clock::local_time();
	vector<TBackupUnit*> punitsAsked;
	size_t i;
	for (i= 0; i<units_.size()-1; ++i){//ǉ̂͏
		if (size<sizeMax_) break;
		size-= units_[i].FilesSize();
		if (timeCur-units_[i].Time()<tmdDeletable_)
			punitsAsked.push_back(&units_[i]);
	}
	size_t ctDeletedUnit= i;
	bool bDeleteRecent= punitsAsked.empty()
		|| onQueryDelete_.empty() || onQueryDelete_(punitsAsked);
	i= 0;
	for (int j= 0; j<ctDeletedUnit; ++j){
		if (bDeleteRecent || timeCur-units_[i].Time()>=tmdDeletable_){
			units_[i].Delete();
			units_.erase(units_.begin()+i);
		}else{
			++i;
		}
	}
}

//---------------------------------------------------------------------------
void TBackup::ReadMembers(gimite::TGTSNode* node)
{
	GIMITE_MAPROFILED("TBackup::ReadMembers");
	node->Read("Dir", pathBaseDir_);
	node->Read("MaxSize", sizeMax_);
	node->Read("DeletableTime", tmdDeletable_);
	units_.clear();
	while (node->ObjectsCount("Unit")){
		units_.push_back(TBackupUnit(this));
		node->ReadObject("Unit", &units_.back());
	}
}

//---------------------------------------------------------------------------
void TBackup::WriteMembers(gimite::TGTSNode* node)const
{
	GIMITE_MAPROFILED("TBackup::WriteMembers");
	node->Write("Dir", pathBaseDir_);
	node->Write("MaxSize", sizeMax_);
	node->Write("DeletableTime", tmdDeletable_);
	for (size_t i= 0; i<units_.size(); ++i)
		node->WriteObject("Unit", &units_[i]);
}

//---------------------------------------------------------------------------
DWORD TBackup::TotalSize()const//SobNAbṽTCY̍v
{
	DWORD size= 0;
	for (size_t i= 0; i<units_.size(); ++i)
		size+= units_[i].FilesSize();
	return size;
}

//---------------------------------------------------------------------------
DWORD TBackup::FilesSize(const TStringVector& paths)
	//w̃t@C̃TCY̍v
{
	DWORD size= 0;
	for (size_t i= 0; i<paths.size(); ++i)
		size+= GetCompressedFileSize(paths[i].c_str(), NULL);
			//4Gȏ̃t@Cɂ͖Ή
	return size;
}

//---------------------------------------------------------------------------

