//---------------------------------------------------------------------------
#ifndef FileSetH
#define FileSetH
//---------------------------------------------------------------------------
#include <set>
#include <map>
#include <boost/function.hpp>
#include "File.h"
#include "SearchedPattern.h"
#include "ReplacingPattern.h"
#include "Exceptions.h"
//---------------------------------------------------------------------------
class TAbstractPlaceList;
template <class T>
class TPlaceList;
class TBackup;
//---------------------------------------------------------------------------
enum TSearchResult{//
	srOk,//
  srInvalid,//񂪕s
  srNoFiles,//Ώۃt@C
	srStopped,//~ꂽ
	srError,//G[
};

enum TReplaceResult{//u
	rrOk,//
	rrStopped,//~ꂽ
};

enum TProcessType{//ߒ̎
	ptEnumFiles,//t@C񋓒
	ptSearch,//
	ptBackup,//obNAbv
	ptReplace,//u
};

enum TDecision{//[Ǔ
	dcRetry,//Ďs
	dcSkip,//XLbv
	dcForce,//s
	dcAbort,//~
};

/*
enum TFileErrorType{//t@CANZXG[̎
	fetCantOpen,//JȂ
*/
//---------------------------------------------------------------------------
typedef boost::function4<void, TProcessType, std::string /*path*/,
	int /*nCurrent*/, int /*ctAll*/> TOnReportProcess;
typedef boost::function2<TDecision, std::string /*path*/,
	TFileOperationResult /*r*/> TOnFileError;
//---------------------------------------------------------------------------
//t@CQ
class TFileSet
{
public:
	typedef std::vector<std::string> TStringVector;

	explicit TFileSet(TBackup* backup): ccode_(ccSjis), bCheckCharCode_(true),
		bCheckTime_(true), sizeMaxFile_(INT_MAX), backup_(backup){}

	void SetCharCode(TCharCode ccode){ccode_= ccode;}//R[hw
	void SetCheckCharCode(bool b){bCheckCharCode_= b;}
		//uɕR[h`FbN邩w
	void SetCheckTime(bool b){bCheckTime_= b;}
		//uɃ^CX^v`FbN邩w
  void SetMaxFileSize(int n){sizeMaxFile_= n;}
    //t@CTCY̏ioCgjw
	void SetFiles(const TStringVector& paths)//Ώۃt@Cw
		{paths_= paths;}
	void SetDir(const std::string& pathDir, const TStringVector& ssFilter,
		bool bSubDir);//ΏۃtH_w

	TSearchResult Search(TAbstractSearchedPattern* pattern,
			TAbstractPlaceList* listPlace);
		//patterngČʂlistPlaceɒǉ
	TReplaceResult Replace(TAbstractReplacingPattern* pattern,
			TAbstractPlaceList* listPlace, bool bBackup);
		//patternglistPlacȇIꂽʒuu

	TFile<char>* NewFile(const std::string& path, char*)const;
		//pathŎw肳ꂽVOoCg̃t@CIuWFNg
	TFile<wchar_t>* NewFile(const std::string& path, wchar_t*)const;
		//pathŎw肳ꂽ}`oCg̃t@CIuWFNg
	bool MultiByte()const{return ccode_!=ccAscii;}//}`oCgň
	bool TargetExists()const{return !paths_.empty();}//Ώۃt@C݂邩
	void SetOnReportProcess(TOnReportProcess func){onReportProcess_= func;}
	void SetOnFileError(TOnFileError func){onFileError_= func;}

private:
	typedef std::set<std::string> TPathSet;
  typedef std::map<std::string, boost::shared_ptr<TPosixTime> > TFileTimes;

	TStringVector paths_;//pX̔z
	TCharCode ccode_;//R[h̎w
	bool bCheckCharCode_;//uɕR[h`FbN邩
  bool bCheckTime_;//uɃ^CX^v`FbN邩
  int sizeMaxFile_;//t@CTCY̏ioCgj
	TBackup* backup_;//obNAbv(UN)
  TFileTimes timesFile_;//iŃqbgjet@C̃^CX^v
	TOnReportProcess onReportProcess_;//ߒ̕
	TOnFileError onFileError_;//t@CANZX̃G[

  template <class T>
  bool SearchFile(const std::string& path, TSearchedPattern<T>* pattern,
	  	TPlaceList<T>* listPlace);
    //path̃t@CpatternAlistPlaceɉ(*2)
  template <class T>
  void ReplaceFile(const std::string& path, TReplacingPattern<T>* pattern,
	  	TPlaceList<T>* listPlace);
    //path̃t@ClistPlaceőIĂʒupatternŒu(*3)
  bool AddFilesInDir(const std::string& pathDir,
  	const TStringVector& ssFilter);//pathDir̃t@CΏۂɒǉ
  bool AddFileToSet(const std::string& path, TPathSet* pset_path)
	  	{pset_path->insert(path); return true;}
		//path*pset_pathɒǉ
	template <class T>
	void ErasePlacesInFile(TPlaceList<T>* listPlace, const std::string& path);
		//w̃t@CɑʒuʒuXg폜
  bool IsFileChanged(const std::string& path);
    //t@CɕύXꂽ
	void ReportProcess(TProcessType type, const std::string& path,
	  	int nCurrent, int ctAll);
		//ߒ(*2)
	TDecision QueryFileError(const std::string& path, TFileOperationResult r);
		//t@CG[̏uu

};

//*1	ssFilter: t@C̃tB^(*.txtȂ)
//		bSubDir: TutH_܂ނ
//*2	qbgtrueԂ
//*3  XLbvꍇfalseԂ
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//ev[g֐̎
//---------------------------------------------------------------------------
template <class T>
bool TFileSet::SearchFile(const std::string& path,
		TSearchedPattern<T>* pattern, TPlaceList<T>* listPlace)
  //path̃t@CpatternAlistPlaceɉ(*2)
{
	GIMITE_MAPROFILED("TFileSet::SearchFile");
  std::auto_ptr<TFile<T> > file(NewFile(path, (T*)0));
  TFileOperationInfo info;
  info.sizeMaxFile_= sizeMaxFile_;
	TFileOperationResult r;
	while (true){
		r= file->LoadText(info);
    if (r==forOk) break;
    else if (r!=forError) return false;
		TDecision dc= QueryFileError(path, r);//G[߂
    switch (dc){
      case dcRetry: continue;
      case dcSkip: return false;
      case dcForce: break;
      case dcAbort: throw EStoppedException();
      default: assert(false);
    }
	}
 	return pattern->Search(file->Text(), listPlace, path);
}

//---------------------------------------------------------------------------
template <class T>
void TFileSet::ReplaceFile(const std::string& path,
		TReplacingPattern<T>* pattern, TPlaceList<T>* listPlace)
	//path̃t@ClistPlaceőIĂʒupatternŒu(*2)
{
	std::auto_ptr<TFile<T> > file(NewFile(path, (T*)0));
	TFileOperationInfo info;
	info.bCheckCharCode_= bCheckCharCode_;
	TFileOperationResult r;
	while (true){
    TDecision dc= dcForce;
    if (IsFileChanged(path)){
      dc= QueryFileError(path, forChanged);
    }
    if (dc==dcForce){
  		if ((r= file->LoadText(info))==forOk) break;
  		dc= QueryFileError(path, r);//G[߂
    }
		if (dc==dcForce) break;
		if (dc==dcAbort) throw EStoppedException();//~
		if (dc==dcSkip) return;
	}
	bool bReplaced= pattern->Replace(file->Text(), listPlace, path);
  if (!bReplaced) return; //1uȂ
	while (true){
		if ((r= file->SaveText())!=forError) break;
		TDecision dc= QueryFileError(path, r);//G[߂
		if (dc==dcRetry) continue;
		return;
	}
	ErasePlacesInFile(listPlace, path);//Ïʒu폜
}

//---------------------------------------------------------------------------
template <class T>
void TFileSet::ErasePlacesInFile(TPlaceList<T>* listPlace,
		const std::string& path)
	//w̃t@CɑʒuʒuXg폜
{
	for (int i= 0; i<listPlace->PlaceCount(); ){
		if (listPlace->Place(i)->Path()==path)
			listPlace->Erase(i);
		else
			++i;
	}
}

//---------------------------------------------------------------------------
#endif
