//---------------------------------------------------------------------------
#ifndef SearchedPatternH
#define SearchedPatternH
//---------------------------------------------------------------------------
#include <string>
#include <boost/regex.hpp>
#include "Definitions.h"
#include "StrProcess.h"
#include "PlaceList.h"
//---------------------------------------------------------------------------
std::istream& operator>>(std::istream& is, TSearchMode& ccode);
std::ostream& operator<<(std::ostream& os, TSearchMode ccode);
//---------------------------------------------------------------------------
struct TSearchInfo//p^[邽߂̏
{
	bool bMultiByte_;//}`oCg
  std::string sSearched_;//
  TSearchMode mode_;//[h
  bool bCaseSensitive_;//啶Eʂ邩
  bool bEscape_;//GXP[vg
  bool bByLine_;//sPʌ
  bool bRecordLine_;//sԍL^邩
};

//---------------------------------------------------------------------------
//p^[̋ʊNX
class TAbstractSearchedPattern
{
public:
	virtual ~TAbstractSearchedPattern(){}
  virtual bool Valid()const= 0;//LȌp^[
	static TAbstractSearchedPattern*
  		NewSearchedPattern(const TSearchInfo& info);
    //w̏𖞂p^[

private:
	template <class T>
  static std::basic_string<T> MakeWordRegex(const std::basic_string<T>& ts);
  	//tsƒPPʂœ̂鐳K\

};

//---------------------------------------------------------------------------
//p^[
template <class TChar>
class TSearchedPattern: public TAbstractSearchedPattern
{
public:
	typedef std::basic_string<TChar> TString;
	typedef std::vector<typename TString::size_type> TSizeVector;

  TSearchedPattern(bool bByLine, bool bRecordLine)
  	: bByLine_(bByLine), bRecordLine_(bRecordLine){}
  bool Search(const TString& text, TPlaceList<TChar>* listPlace,
	 		const std::string& path);
  	//textp^[AlistPlaceɒǉ(*1)

protected:
	typedef TPlaceList<TChar>::TReferenceStrings TReferenceStrings;

	struct TPlaceAddingInfo{//ʒuǉ̂߂̕t
  	TPlaceList<TChar>* listPlace_;//ʒuXg
  	std::string path_;//pX
    int nLine_;//sԍ
    TSizeVector posesReturn_;//sʒu̔z
  };

	virtual bool SearchLine(const TString& text,
  		const TPlaceAddingInfo* info)= 0;
  	//1stextp^[(*1)(*2)
  void AddPlaceToList(const TPlaceAddingInfo* info, const TString& text,
	  	int pos, int len,
      const TReferenceStrings& ssRef= TReferenceStrings());
  	//ʒuXgɒǉ

private:
	bool bByLine_;//sPʌ
  bool bRecordLine_;//sԍL^邩

};

//*1	qbgtrueԂB
//*2	Ŝ1sƂ݂ȂƂLB
//---------------------------------------------------------------------------
//ʌ̌p^[
template <class TChar>
class TNormalSearchedPattern: public TSearchedPattern<TChar>
{
public:
	typedef std::basic_string<TChar> TString;

  explicit TNormalSearchedPattern(const TString& ts, bool bByLine,
  		bool bRecordLine)
  	: TSearchedPattern<TChar>(bByLine, bRecordLine), tsSearched_(ts){}
  virtual bool Valid()const{return !tsSearched_.empty();}
  	//LȌp^[

protected:
	typedef TSearchedPattern<TChar>::TPlaceAddingInfo TPlaceAddingInfo;

	virtual bool SearchLine(const TString& text,
  		const TPlaceAddingInfo* info);
  	//1stextp^[

private:
	TString tsSearched_;//

};

//---------------------------------------------------------------------------
//K\̌p^[
template <class TChar>
class TRegexSearchedPattern: public TSearchedPattern<TChar>
{
public:
	typedef std::basic_string<TChar> TString;

  TRegexSearchedPattern(const TString& tsRegex,
  	boost::regex::flag_type flagRegex, bool bByLine, bool bRecordLine);
  virtual bool Valid()const{return bValid_;}
  	//LȌp^[

protected:
	typedef TSearchedPattern<TChar>::TPlaceAddingInfo TPlaceAddingInfo;

	virtual bool SearchLine(const TString& text,
  		const TPlaceAddingInfo* info);
  	//1stextp^[

private:
	typedef boost::match_results<typename TString::const_iterator> TResult;

	struct TGrepPredicate{//reg_grepĂяo֐
		bool operator()(const TResult& r);
    TRegexSearchedPattern* pattern_;
    const TPlaceAddingInfo* info_;//ʒuǉ̂߂̏
    const TString* ptext_;//Ώۂ̃eLXg
    TGrepPredicate(): pattern_(NULL), info_(NULL), ptext_(NULL){}
	};
  friend struct TGrepPredicate;

	boost::reg_expression<TChar> regex_;//K\
  bool bValid_;//LȐK\
  bool bStopped_;//~ꂽ

};

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//ev[g֐̎
//---------------------------------------------------------------------------
template <class T>
std::basic_string<T> TAbstractSearchedPattern::
		MakeWordRegex(const std::basic_string<T>& ts)
  //tsƒPPʂœ̂鐳K\
{
  typedef std::basic_string<T>::size_type TSize;
		std::basic_string<T> tsR;
	TSize pos= 0;
	while (pos<ts.length() && std::isspace(ts[pos])) ++pos;
	if (pos>=ts.length()) return std::basic_string<T>();
	TSize posPrev= pos;
	bool bPrevMark= !IsCAlnum(ts[pos]);
	while (true){
		++pos;
		if (bPrevMark || pos>=ts.length() || !IsCAlnum(ts[pos])){
			if (!tsR.empty()) tsR+= ToTstr<T>("\\s*");
			std::basic_string<T> tsT
				= EscapeForRegex(ts.substr(posPrev, pos-posPrev));
			if (bPrevMark) tsR+= tsT;
			else tsR+= ToTstr<T>("\\<")+tsT+ToTstr<T>("\\>");
			while (pos<ts.length() && std::isspace(ts[pos])) ++pos;
			posPrev= pos;
			if (pos>=ts.length()) break;
		}
		bPrevMark= !IsCAlnum(ts[pos]);
	}
  return tsR;
}

//---------------------------------------------------------------------------
template <class T>
bool TSearchedPattern<T>::
		Search(const TString& text, TPlaceList<T>* listPlace,
    	const std::string& path)
  //textp^[AlistPlaceɒǉ
{
	GIMITE_MAPROFILED("TSearchedPattern<T>::Search");
  bool bHit= false;
  TPlaceAddingInfo info;
  info.listPlace_= listPlace;
  info.path_= path;
  info.nLine_= 0;
  if (bByLine_){//sPʌ
  	TString::size_type pos= 0, posPrev= 0;
    while (true){
    	pos= text.find(T(uchar('\n')), pos);
      if (pos==TString::npos) pos= text.length();
      if (SearchLine(text.substr(posPrev, pos-posPrev), &info))
      	bHit= true;
      if (pos>=text.length()) break;
      ++pos;
      posPrev= pos;
      ++info.nLine_;
    }
    return bHit;
  }else{
  	if (bRecordLine_){
			GIMITE_MAPROFILED("TSearchedPattern<T>::Search#return_pos");
	  	TString::size_type pos= 0;
    	while ((pos= text.find(T(uchar('\n')), pos))!=TString::npos){
      	info.posesReturn_.push_back(pos);//sʒuXg
        ++pos;
      }
    }
    return SearchLine(text, &info);
  }
}

//---------------------------------------------------------------------------
template <class T>
void TSearchedPattern<T>::
		AddPlaceToList(const TPlaceAddingInfo* info, const TString& text,
    	int pos, int len, const TReferenceStrings& ssRef)
  //ʒuXgɒǉ
{
  assert(info && info->listPlace_);
  info->listPlace_->Add(info->path_, info->nLine_, text, pos, len,
  	ssRef, info->posesReturn_);
}

//---------------------------------------------------------------------------
template <class T>
bool TNormalSearchedPattern<T>::
		SearchLine(const TString& text, const TPlaceAddingInfo* info)
  //1stextp^[
{
  bool bHit= false;
  TString::size_type pos= 0;
  TString::size_type len= tsSearched_.length();
  while (true){
  	pos= text.find(tsSearched_, pos);
    if (pos==TString::npos) break;
    AddPlaceToList(info, text, pos, len);
    pos+= len;//d͔F߂Ȃ
    bHit= true;
  }
  return bHit;
}

//---------------------------------------------------------------------------
template <class T>
TRegexSearchedPattern<T>::
	TRegexSearchedPattern(const TString& tsRegex,
  		boost::regex::flag_type flagRegex, bool bByLine, bool bRecordLine)
    : TSearchedPattern<T>(bByLine, bRecordLine), bValid_(true)
{
	try{
  	regex_.assign(tsRegex.c_str(), flagRegex);
  }catch (boost::bad_expression&){
  	bValid_= false;
  }
}

//---------------------------------------------------------------------------
template <class T>
bool TRegexSearchedPattern<T>::
		SearchLine(const TString& text, const TPlaceAddingInfo* info)
  //1stextp^[
{
	TGrepPredicate pred;
  pred.pattern_= this;
  pred.info_= info;
	pred.ptext_= &text;
  return boost::regex_grep(pred, text, regex_);//grep
}

//---------------------------------------------------------------------------
template <class T>
bool TRegexSearchedPattern<T>::TGrepPredicate::operator()(const TResult& r)
{
  TPlaceList<T>::TReferenceStrings ssRef;
  for (int i= 0; i<r.size(); ++i)
  	ssRef.push_back(r[i]);
  pattern_->AddPlaceToList(info_, *ptext_, r.position(), r.length(), ssRef);
  return true;
}

//---------------------------------------------------------------------------
#endif
