//---------------------------------------------------------------------------
/*
               蓮vt@C Maprof 1.0  by Gimite s
*/
//---------------------------------------------------------------------------
#ifndef GIMITE_MaprofH
#define GIMITE_MaprofH
//---------------------------------------------------------------------------
#ifdef GIMITE_MAPROF
//---------------------------------------------------------------------------
#include <deque>
#include <map>
#include <vector>
#include <cstring>
#include <string>
#include <ctime>
#include <gimite/singleton_ptr.h>
#ifdef GIMITE_MAPROF_MULTI_THREAD
#include <gimite/thread_local_ptr.h>
#endif
//---------------------------------------------------------------------------
namespace gimite{
//---------------------------------------------------------------------------
class TMaprof
{
public:
	TMaprof(): sFile_("maprof"), timeBegin_(Clock()){}
  ~TMaprof();
	static TMaprof* Instance();
	static void Begin(const char* csName, bool bOperation= false){
		Instance()->events_.push_back(
			TEvent(Clock(), etBegin, csName, bOperation));
	}
	static void End(const char* csName){
		Instance()->events_.push_back(TEvent(Clock(), etEnd, csName, false));
	}
	static void SetFilePrefix(const char* csFile){
		Instance()->sFile_= csFile;
	}

private:
  enum TEventType{etBegin, etEnd};
  struct TEvent{
    double time_;
    TEventType type_;
    const char* csName_;
    bool bOperation_;
    TEvent(){}
    TEvent(double tm, TEventType type, const char* cs, bool b)
      : time_(tm), type_(type), csName_(cs), bOperation_(b){}
	};
	typedef std::deque<TEvent> TEvents;

  struct TMaprofiledInfo{
    const char* csName_;
    double timeCost_;
    double timeAll_;
    int ctCalled_;
    explicit TMaprofiledInfo(const char* cs= 0): csName_(cs), timeCost_(0.0),
      timeAll_(0.0), ctCalled_(0){}
    bool operator<(const TMaprofiledInfo& info2)const{
      return timeCost_<info2.timeCost_;
    }
  };
  struct TCall{
    const char* csName_;
    double timeBegin_;
    double timeBlockBegin_;
    double timeCost_;
    bool bOperation_;
    TCall(const char* cs, double tm, bool b): csName_(cs), timeBegin_(tm),
      timeBlockBegin_(tm), timeCost_(0.0), bOperation_(b){}
  };

  typedef std::map<std::string, TMaprofiledInfo> TInfoMap;
  typedef std::map<std::string, TInfoMap> TInfoMapMap;
  typedef std::vector<TMaprofiledInfo> TInfos;

  std::string sFile_;
  TEvents events_;
  double timeBegin_;

  //TInfoMapɏo͂
  void PrintInfos(std::ostream& os, const TInfoMap& mp, double timeWhole,
    int lenMaxName);
  //ݎ[b]
  static double Clock(){ return double(std::clock())/CLK_TCK; }

};

#ifdef GIMITE_MAPROF_MULTI_THREAD
static thread_singleton_ptr<TMaprof> maprof;
#else
static singleton_ptr<TMaprof> maprof;
#endif
//---------------------------------------------------------------------------
class TMaprofiled
{
public:
	explicit TMaprofiled(const char* csName, bool bOperation= false)
			:csName_(csName){
		TMaprof::Begin(csName_, bOperation);
	}
	~TMaprofiled(){
		TMaprof::End(csName_);
  }

private:
	const char* csName_;

};

//---------------------------------------------------------------------------
inline TMaprof* TMaprof::Instance()
{
#ifdef GIMITE_MAPROF_MULTI_THREAD
	if (!maprof.get()) maprof.reset(new TMaprof);
#endif
	return maprof.get();
}

//---------------------------------------------------------------------------
}//namespace gimite
//---------------------------------------------------------------------------
# define GIMITE_MAPROFILED(a) gimite::TMaprofiled gimite_maprofiled(a)
# define GIMITE_MAPROFILED_OP(a) gimite::TMaprofiled gimite_maprofiled(a, true)
# define GIMITE_MAPROF_BEGIN(a) (gimite::TMaprof::Begin(a))
# define GIMITE_MAPROF_END(a) (gimite::TMaprof::End(a))
# define GIMITE_MAPROF_BEGIN_OP(a) (gimite::TMaprof::Begin(a, true))
# define GIMITE_MAPROF_END_OP(a) (gimite::TMaprof::End(a))
# define GIMITE_MAPROF_SET_FILE_PREFIX(a) (gimite::TMaprof::SetFilePrefix(a))

#else

# define GIMITE_MAPROFILED(a)
# define GIMITE_MAPROFILED_OP(a)
# define GIMITE_MAPROF_BEGIN(a)
# define GIMITE_MAPROF_END(a)
# define GIMITE_MAPROF_BEGIN_OP(a)
# define GIMITE_MAPROF_END_OP(a)
# define GIMITE_MAPROF_SET_FILE_PREFIX(a)

#endif
//---------------------------------------------------------------------------
#endif
